urls = []


def route(url, **kwargs):

    def handler_wrapper(handler):
        urls.append((url, handler, kwargs))
        return handler
    return handler_wrapper


def routes():
    return urls




import operator
from functools import reduce


def merge_args(*args, **kwargs):
    _kwargs = [(k, v) for k, v in kwargs.items() if v is not None]
    _kwargs = _kwargs if len(_kwargs) > 0 else [()]
    return operator.add(args, reduce(operator.add, _kwargs))

def chunks(l, n=2):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]



import asyncio
from time import time
from .util import merge_args
from .module import IMModule


class IM(IMModule):

    def __init__(self, client):
        self.client = client
        self.execute_command = client.execute_command
        self.recive_script = self.client.register_script(
            """
            local key = KEYS[1]
            local res = redis.call("XRANGE", KEYS[1], ARGV[1], ARGV[1])
            if res ~= nil and #res == 1 then
                return res[1]
            end
            return nil
            """
        )
        self.message_script = self.client.register_script(
            """
            local key = KEYS[1]
            local res = redis.call("XRANGE", KEYS[1], ARGV[1], ARGV[1])
            if res ~= nil and #res == 1 then
                return res[1]
            end
            return nil
            """
        )
        self.send_script = self.client.register_script(
            """
            local s_to_user_id = KEYS[1]
            local s_user_id = KEYS[2]
            local r_to_user_id = KEYS[3]
            local user_id = table.remove(ARGV, 1)
            local mid = redis.call("XADD", s_to_user_id, "*", "uid", user_id, unpack(ARGV))
            redis.call("XADD", s_user_id, mid, "FW", s_to_user_id)
            redis.call("ZADD", r_to_user_id, "INCR", 1, s_user_id)
            return mid
            """
        )
        self.gsend_script = self.client.register_script(
            # 这里只能确定两个key, 剩下的key是从smembers读取的
            """
            local s_group_id = KEYS[1]
            local m_group_id = KEYS[2]
            local s_user_id = table.remove(ARGV, 1)
            local mid = redis.call("XADD", s_group_id, "*", "uid", s_user_id, unpack(ARGV))
            local members = redis.call("SMEMBERS", m_group_id)
            for i, uid in ipairs(members) do
                redis.call("XADD", "s:" .. uid, mid, "FW", s_group_id)
                if uid ~= s_user_id then
                    redis.call("ZADD", "r:" .. uid, "INCR", 1, s_group_id)
                end
            end
            return mid
            """
        )
        self.join_script = self.client.register_script(
            """
            local c_user_id = KEYS[1]
            local r_user_id = KEYS[2]
            local m_group_id = KEYS[3]
            local s_group_id = KEYS[4]
            local user_id = ARGV[1]
            local group_id = ARGV[2]
            local timestamp, ms = unpack(redis.call("TIME"))
            -- 1. add user contact;
            redis.call("ZADD", c_user_id, timestamp .. "000", group_id)
            -- 2. add group member;
            redis.call("SADD", m_group_id, user_id)
            -- 3. send join group message;
            local mid = redis.call("XADD", s_group_id, "*", "action", "join", "uid", user_id, "gid", group_id)
            local members = redis.call("SMEMBERS", m_group_id)
            for i, uid in ipairs(members) do
                redis.call("XADD", "s:" .. uid, mid, "FW", s_group_id)
                if uid ~= user_id then
                    redis.call("ZADD", "r:" .. uid, "INCR", 1, s_group_id)
                end
            end
            """
        )
        self.quit_script = self.client.register_script(
            """
            local c_user_id = KEYS[1]
            local r_user_id = KEYS[2]
            local m_group_id = KEYS[3]
            local s_group_id = KEYS[4]
            local user_id = ARGV[1]
            local group_id = ARGV[2]
            -- 1. remove user contact;
            redis.call("ZREM", c_user_id, group_id)
            redis.call("ZREM", r_user_id, "s:" .. group_id)
            -- 2. remove user from group
            local res = redis.call("SREM", m_group_id, user_id)
            -- 3. send quit group message
            local mid = redis.call("XADD", s_group_id, "*", "action", "quit", "uid", user_id, "gid", group_id)
            local members = redis.call("SMEMBERS", m_group_id)
            for i, uid in ipairs(members) do
                redis.call("XADD", "s:" .. uid, mid, "FW", s_group_id)
            end
            return res
            """
        )
        self.group_info_script = self.client.register_script(
            """
            local i_group_id = KEYS[1]
            local m_group_id = KEYS[2]
            local info = redis.call("HGETALL", i_group_id)
            local members = redis.call("SMEMBERS", m_group_id)
            return { info, members }
            """
        )
        self.create_group_script = self.client.register_script(
            # 1. hmset group info; 2. join_script logic
            """
            local i_group_id = KEYS[1]
            local c_user_id = KEYS[2]
            local r_user_id = KEYS[3]
            local m_group_id = KEYS[4]
            local s_group_id = KEYS[5]
            local group_id = table.remove(ARGV, 1)
            local user_id = table.remove(ARGV, 1)
            local timestamp, ms = unpack(redis.call("TIME"))
            -- 1. hmset group info
            redis.call("HMSET", i_group_id, "master", user_id, unpack(ARGV))
            -- 2. add user contact;
            redis.call("ZADD", c_user_id, timestamp .. "000", group_id)
            -- 3. add group member;
            redis.call("SADD", m_group_id, user_id)
            -- 4. send join group message;
            local mid = redis.call("XADD", s_group_id, "*", "action", "join", "uid", user_id, "gid", group_id)
            local members = redis.call("SMEMBERS", m_group_id)
            for i, uid in ipairs(members) do
                redis.call("XADD", "s:" .. uid, mid, "FW", s_group_id)
                if uid ~= user_id then
                    redis.call("ZADD", "r:" .. uid, "INCR", 1, s_group_id)
                end
            end
            """
        )
        self.link_script = self.client.register_script(
            """
            local s_to_user_id = KEYS[1]
            local s_user_id = KEYS[2]
            local r_to_user_id = KEYS[3]
            local c_user_id = KEYS[4]
            local user_id = ARGV[1]
            local to_user_id = ARGV[2]
            local timestamp, ms = unpack(redis.call("TIME"))
            local res = redis.call("ZADD", c_user_id, timestamp, s_to_user_id)
            if res ~= 0 then
                local mid = redis.call("XADD", s_to_user_id, "*", "action", "link", "uid", user_id, "tuid", to_user_id)
                redis.call("XADD", s_user_id, mid, "FW", s_to_user_id)
                redis.call("ZADD", r_to_user_id, "INCR", 1, s_user_id)
            end
            return res
            """
        )
        self.unlink_script = self.client.register_script(
            """
            local c_user_id = KEYS[1]
            local r_user_id = KEYS[2]
            local to_user_id = ARGV[1]
            redis.call("ZREM", r_user_id, "s:" .. to_user_id)
            return redis.call("ZREM", c_user_id, to_user_id)
            """
        )

    def recive(self, user_id, block=10000, count=20, start=0):
        res = self.execute_command("XREAD", "COUNT", count, "BLOCK", block, "STREAMS", f"s:{user_id}", start)
        return res if isinstance(res, list) and len(res) > 0 else None

    def message(self, key, message_id, group=False):
        return self.message_script(
            keys=[f"gs:{key}" if group else f"s:{key}"],
            args=[message_id]
        )

    def send(self, user_id, to_user_id, *args, **kwargs):
        return self.send_script(
            keys=[f"s:{to_user_id}", f"s:{user_id}", f"r:{to_user_id}"],
            args=merge_args(user_id, *args, **kwargs)
        )

    def gsend(self, user_id, group_id, *args, **kwargs):
        return self.gsend_script(
            keys=[f"gs:{group_id}", f"gm:{group_id}"],
            args=merge_args(user_id, *args, **kwargs)
        )

    def user(self, user_id, **kwargs):
        if kwargs:
            return self.execute_command(*merge_args("HMSET", f"u:{user_id}", **kwargs))
        return self.execute_command("HGETALL", f"u:{user_id}")

    def group(self, group_id, user_id=None, **kwargs):
        if kwargs:
            return self.create_group_script(
                keys=[f"gi:{group_id}", f"c:{user_id}", f"r:{user_id}", f"gm:{group_id}", f"gs:{group_id}"],
                args=merge_args(group_id, user_id, **kwargs)
            )
        return self.group_info_script(keys=[f"gi:{group_id}", f"gm:{group_id}"], args=[])

    def link(self, user_id, to_user_id):
        return self.link_script(
            keys=[f"s:{to_user_id}", f"s:{user_id}", f"r:{to_user_id}", f"c:{user_id}"],
            args=[user_id, to_user_id]
        )

    def unlink(self, user_id, to_user_id):
        return self.unlink_script(
            keys=[f"c:{user_id}", f"r:{user_id}"],
            args=[to_user_id]
        )

    def join(self, user_id, group_id):
        return self.join_script(
            keys=[f"c:{user_id}", f"r:{user_id}", f"gm:{group_id}", f"gs:{group_id}"],
            args=[user_id, group_id]
        )

    def quit(self, user_id, group_id):
        return self.quit_script(
            keys=[f"c:{user_id}", f"r:{user_id}", f"gm:{group_id}", f"gs:{group_id}"],
            args=[user_id, group_id]
        )


class AsyncIM(IM):

    async def recive(self, user_id, block=10000, count=20, start=0):
        res = await self.execute_command("XREAD", "COUNT", count, "BLOCK", block, "STREAMS", f"s:{user_id}", start)
        return res if isinstance(res, list) and len(res) > 0 else None




from .util import merge_args, chunks


def parse_recive_response(response):
    return [[response[0][0], [(mid, dict(chunks(message))) for mid, message in response[0][1]] ]] if response else None



class IMModule:

    def __init__(self, client):
        self.client = client
        self.execute_command = client.execute_command
        self.client.set_response_callback('IM.RECIVE', parse_recive_response)

    def recive(self, user_id, block=10000, count=20, start=0):
        params = merge_args("IM.RECIVE", user_id, BLOCK=block, COUNT=count, START=start)
        return self.execute_command(*params)

    def message(self, key, message_id, group=False):
        params = ["IM.MESSAGE", "GROUP" if group else "USER", key, message_id]
        return self.execute_command(*params)

    def send(self, user_id, to_user_id, *args, **kwargs):
        params = merge_args("IM.SEND", user_id, to_user_id, *args, **kwargs)
        return self.execute_command(*params)

    def gsend(self, user_id, group_id, *args, **kwargs):
        params = merge_args("IM.GSEND", user_id, group_id, *args, **kwargs)
        return self.execute_command(*params)

    def user(self, user_id, **kwargs):
        params = merge_args("IM.USER", user_id, **kwargs)
        return self.execute_command(*params)

    def group(self, group_id, user_id=None, **kwargs):
        params = merge_args("IM.GROUP", group_id, user_id=user_id, **kwargs)
        return self.execute_command(*params)

    def link(self, user_id, to_user_id):
        params = merge_args("IM.LINK", user_id, to_user_id)
        return self.execute_command(*params)

    def unlink(self, user_id, to_user_id):
        params = merge_args("IM.UNLINK", user_id, to_user_id)
        return self.execute_command(*params)

    def join(self, user_id, group_id):
        params = merge_args("IM.JOIN", user_id, group_id)
        return self.execute_command(*params)

    def quit(self, user_id, group_id):
        params = merge_args("IM.QUIT", user_id, group_id)
        return self.execute_command(*params)



from redis.commands import RedisModuleCommands, AsyncRedisModuleCommands
from .module import *
from .im import *


def im(self, module=True):
    return IMModule(self) if module else IM(self)

def asyncim(self, module=True):
    return IMModule(self) if module else AsyncIM(self)


RedisModuleCommands.im = im
AsyncRedisModuleCommands.im = asyncim


# try:
#     from redis.exceptions import NoScriptError
#     from redis.asyncio.connection import DefaultParser as AsyncDefaultParser
#     from redis.connection import DefaultParser
#     # using kvrocks instead of redis
#     AsyncDefaultParser.EXCEPTION_CLASSES['ERR']['NOSCRIPT No matching script. Please use EVAL'] = NoScriptError
#     DefaultParser.EXCEPTION_CLASSES['ERR']['NOSCRIPT No matching script. Please use EVAL'] = NoScriptError
# except Exception:
#     pass


__version__ = "0.0.4"



from os.path import dirname, join, isfile
from tornado.options import define, parse_config_file, parse_command_line, options


define('DEBUG', default=True)
define('SERVER_PORT', default=8888)

define('REDIS_HOST', default="localhost")
define('REDIS_PORT', default=6379)

root_path = dirname(dirname(__file__))
config_file = join(root_path, "etc", "web_config.conf")

define(
    "CONFIG",
    help="tornado config file",
    default=config_file,
    callback=lambda path: isfile(path) and parse_config_file(path, final=False),
)



import re

from setuptools import find_namespace_packages, setup

with open("redisim/__init__.py") as f:
    version = re.findall("__version__.*(\d+\.\d+\.\d+).*", f.read())[0]

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="redisim",  # package name
    version=version,  # package version
    description="redisim",  # package description
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": "http://github.com/lloydzhou/redisim",
        "Code": "http://github.com/lloydzhou/redisim",
        "Issue tracker": "http://github.com/lloydzhou/redisim/issues",
    },
    author="lloydzhou@gmail.com",
    license="MIT",
    keywords=["Redis", "IM", "Websocket"],
    packages=find_namespace_packages(),
    zip_safe=False,
    install_requires=["redis>=4.2.0"],
    python_requires=">=3.8",
)



import asyncio
import logging
import json
import redisim
import redis.asyncio as redis
# import redis

import tornado.log
import tornado.web
import tornado.ioloop
import tornado.autoreload
import tornado.httpserver
import tornado.websocket
from os.path import abspath, dirname, join
from config import options, parse_command_line

from core.route import route, routes


redis_url = "redis://%s:%s" % (options.REDIS_HOST, options.REDIS_PORT)
pool = redis.ConnectionPool.from_url(redis_url, decode_responses=True)
client = redis.Redis(connection_pool=pool)
# module=False, using lua script
# module=True, using redis module
im = client.im(module=False)


def array_to_dict(args, **kwargs):
    if isinstance(args, (list, tuple)):
        kwargs.update(dict(args[i:i+2] for i in range(0, len(args), 2)))
    elif isinstance(args, dict):
        kwargs.update(args)
    return kwargs


async def login(user_id, passwd=None, *args, **kwargs):
    await im.user(user_id, **kwargs)
    res = await im.user(user_id)
    return array_to_dict(res, uid=user_id)


async def recive(user_id, **kwargs):
    # return mid, uid, tuid, gid, message
    res = await im.recive(user_id, **kwargs)
    if isinstance(res, list):
        channel, messages = res[0]
        # print('channel', channel)
        for mid, message in messages:
            m = array_to_dict(message, tuid=channel[2:])
            if 'FW' in m:
                fwt, fwc = m['FW'].split(':')
                mr = await im.message(fwc, mid, group=fwt == 'gs')
                if mr:
                    mid, message = mr
                    if 'gs' == fwt:
                        m = array_to_dict(message)
                    else:
                        m = array_to_dict(message)
                    yield mid, m['uid'], fwc, fwc if 'gs' == fwt else '', m
            else:
                yield mid, m['uid'], channel[2:], '', m
    yield None, None, None, None, None
    await asyncio.sleep(0.1)


async def im_action(command, *params):
    return await getattr(im, command)(*params)


parse_command_line()
tornado.log.enable_pretty_logging()


@route(r"/ws")
class WebSocketHandler(tornado.websocket.WebSocketHandler):

    async def open(self, *args, **kwargs):
        # print("WebSocket opened", self, self.request, self.request.headers, args, kwargs)
        user_id = self.get_argument('user_id')

        user = await login(user_id, name='user1')

        self.write_message(json.dumps(user))

        async def loop():
            start = self.get_argument('last_message_id') or 0
            while self.ws_connection and not self.ws_connection.is_closing():
                async for mid, uid, tuid, gid, message in recive(user_id, block=10000, count=10, start=start):
                    if mid and mid != start:
                        start = mid
                        # print(mid, message)
                        self.write_message({
                            'id': mid,
                            'uid': uid,
                            'tuid': tuid,
                            'gid': gid,
                            'message': message,
                        })
            # print(mid)
        # do not block cpu
        asyncio.ensure_future(loop())

    async def on_message(self, message):
        # print("WebSocket message", message)
        # self.write_message(u"You said: " + message)
        try:
            message = json.loads(message)
            if 'action' in message:
                action, id, params = message['action'], message.get('id'), message.get('params', [])
                if action.upper() in ['SEND', 'GSEND', 'USER', 'GROUP', 'LINK', 'UNLINK', 'JOIN', 'QUIT']:
                    response = await im_action(action, *params)
                    self.write_message({'id': id, 'response': response})
                    # print('write_message', id, response)
        except Exception as e:
            logging.error('error parse message %s %r', message, e)

    def on_close(self, *args, **kwargs):
        print("WebSocket closed", self, args, kwargs)


if __name__ == "__main__":
    # from modules import *
    root_path = dirname(dirname(abspath(__file__)))
    path = join(root_path, "web", "dist")

    @route(r"/(chat|mine|faxian|address|login)(.*)")
    class RedirectHandler(tornado.web.RequestHandler):
        def get(self, *args):
            self.redirect('/')

    @route(r"/(.*)", path=path, default_filename="index.html")
    class StaticFileHandler(tornado.web.StaticFileHandler):
        pass

    application = tornado.web.Application(routes(), **dict(
        debug=options.DEBUG,
    ))
    server = tornado.httpserver.HTTPServer(application, max_buffer_size=1048576000)
    server.listen(port=options.SERVER_PORT)
    server.start()
    logging.info("Start Success: 0.0.0.0:{}".format(options.SERVER_PORT))

    tornado.ioloop.IOLoop.instance().start()




from redis.commands import RedisModuleCommands, AsyncRedisModuleCommands
from .module import *
from .im import *


def im(self, module=True):
    return IMModule(self) if module else IM(self)

def asyncim(self, module=True):
    return IMModule(self) if module else AsyncIM(self)


RedisModuleCommands.im = im
AsyncRedisModuleCommands.im = asyncim


try:
    from redis.exceptions import NoScriptError
    from redis.asyncio.connection import DefaultParser as AsyncDefaultParser
    from redis.connection import DefaultParser
    # using kvrocks instead of redis
    AsyncDefaultParser.EXCEPTION_CLASSES['ERR']['NOSCRIPT No matching script. Please use EVAL'] = NoScriptError
    DefaultParser.EXCEPTION_CLASSES['ERR']['NOSCRIPT No matching script. Please use EVAL'] = NoScriptError
except Exception:
    pass


__version__ = "0.0.5"
